                                                                        
/* 

Copyright Mathias Paulsen (matpaulsen@gmx.de) 2010  

THERE IS NO WARRANTY FOR THE PROGRAM

This free software; you can redistribute it and/or modify it under the 
the terms of the General Public License as published by the Free Software Foundation; either version 3 of the License, or any later version.

You should have received a copy of the GNU General Public
License along with this GNU Software; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA. 

*/

 
#include <linux/kernel.h>   
#include <linux/module.h>


#include <linux/fs.h>       


#include <asm/uaccess.h>  // needed by put_user 
#include <asm/io.h>       // required for low level io

#include <linux/delay.h> // needed by udelay



static int device_open_counter = 0 ; // how many tlc's are in use ?

                         // io_ports

static int iottyS0 = 0 ; // ttyS0 
static int iottyS1 = 0 ; // ttyS1
static int iottyS2 = 0 ; // ttyS2
static int iottyS3 = 0 ; // ttyS3
static int iolp0   = 0 ; // lp0
static int iolp1   = 0 ; // lp1

// see TLC549 datatsheet
// dimension is s
 
#define CLKhigh  1  
#define CShigh   2 
#define Clow     0
#define BitW    16


/* Setup Time (CS low before first I/O-Clock) */
#define TSU  2
/* I/O-Clock transition time */
#define TTIO 1 
/* duration of CS-High while conversion */
#define TWH 20

module_param( iottyS0  , int  , 0444 );
module_param( iottyS1  , int  , 0444 ); 
module_param( iottyS2  , int  , 0444 ); 
module_param( iottyS3  , int  , 0444 ); 
module_param( iolp0    , int  , 0444 ); 
module_param( iolp1    , int  , 0444 ); 

#define SUCCESS 0


#define DEVICE_NAME "ad_tlc549" // device name

static int Device_Open1 = 0 ; // we don't want to open the device twice
static int Device_Open2 = 0 ; // we want to handle 6 minor numbers
static int Device_Open3 = 0 ;
static int Device_Open4 = 0 ;
static int Device_Open5 = 0 ;
static int Device_Open6 = 0 ;

// handles a device_open

static int device_open(struct inode *inode,
                       struct file *file)
{

  int Device_Open ;
  int minor = inode->i_rdev & 0xff ;
  int io_port ;
  
  switch( minor ) {

  case 1: 
    Device_Open = Device_Open1 ;
    io_port = iottyS0 ;
    break ;
  case 2:
    Device_Open = Device_Open2 ;
    io_port = iottyS1 ;
    break ;
  case 3:
    Device_Open = Device_Open3 ;
    io_port = iottyS2 ;
    break ;
  case 4:
    Device_Open = Device_Open4 ;
    io_port = iottyS3 ;
    break ;
  case 5:  
    Device_Open = Device_Open5 ;
    io_port = iolp0 ;
    break ;
  case 6: 
    Device_Open = Device_Open6 ;
    io_port = iolp1 ;
    break ;
  default: 
    Device_Open = 1 ; 
  }

  if (Device_Open)
    return -EBUSY ; // the device is still open

  if( device_open_counter == 0 ) try_module_get(THIS_MODULE) ; // Macro: prevents unloading a module still in use

  device_open_counter++ ; // increments the number of open devices

  switch( minor )
    {

    case 1:
      Device_Open1++ ; // remember that the device is open
      outb( 64 , iottyS0 + 3 ) ; // turns the tlc549 on
      break ;
    case 2:
      Device_Open2++ ;
      outb( 64 , iottyS1 + 3 ) ;
      break ;
    case 3:
      Device_Open3++ ;
      outb( 64 , iottyS2 + 3 ) ;
      break ;
    case 4:
      Device_Open4++ ;
      outb( 64 , iottyS3 + 3 ) ;
      break ;  
    case 5:
      Device_Open5++ ;
      outb( 128 , iolp0 ) ;
      break ;
    case 6:
      Device_Open6++ ;
      outb( 128 , iolp1 ) ;
      break ;
      
    }
  
  return SUCCESS ;

}

// handles a devie_close


static int device_release(struct inode *inode,
                          struct file *file)
{

  unsigned int minor = inode->i_rdev & 0xff ; // get the minor number

  switch( minor ) {

  case 1:
    Device_Open1-- ;
    outb( 0 , iottyS0 + 3 ) ;
    device_open_counter-- ;
    break ;
  case 2:
    Device_Open2-- ;
    outb( 0 , iottyS1 + 3 ) ;
    device_open_counter-- ;
    break ;
  case 3:
    Device_Open3-- ;
    outb( 0 , iottyS2 + 3 ) ;
    device_open_counter-- ;
    break ;
  case 4:
    Device_Open4-- ;
    outb( 0 , iottyS3 + 3 ) ;
    device_open_counter-- ;
    break ;
  case 5:
    Device_Open5-- ;
    outb( 0 , iolp0 ) ;
    device_open_counter-- ;
    break ;
  case 6:
    Device_Open6-- ;
    outb( 0 , iolp1 ) ;
    device_open_counter-- ;
    break ;
  
  }
  

  if( device_open_counter == 0 ) module_put(THIS_MODULE) ; // now we are able to unload the module


  return 0 ; // see declaration

}



static ssize_t device_read(struct file *file,
    char *buffer,    // The buffer to fill with data
    size_t length,   // The length of the buffer (not used)
    loff_t *offset)  // Our offset in the file 

{

  
  int bytes_read = 0 ; // we have to return the number of Bytes read

  int EV = 0 ;   // endvalue
  char OutBuf[80] ; // we copy a string to the user_space
  char *OutBufPtr ; // a pointer to the string
  int i ; // as usual
  int BV = 128 ; // bitvalue 

  int Input = 0 ;
  int Output = 0 ;


  unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev) ;


  switch( minor ) {
    
  case 1:
    Input = iottyS0 + 6 ;
    Output = iottyS0 + 4 ;
    break ;
  case 2:
    Input = iottyS1 + 6 ;
    Output = iottyS1 + 4 ;
    break ;
  case 3:
    Input = iottyS2 + 6 ;
    Output = iottyS2 + 4 ;
    break ;
  case 4:
    Input = iottyS3 + 6 ;
    Output = iottyS3 + 4 ;
    break ;
  case 5:
    Input = iolp0 + 1 ;
    Output = iolp0 ;
    break ;
  case 6:
    Input = iolp1 + 1 ;
    Output = iolp1 ;
    break ;

  }

  // now we communicate with the tcl549 
 
  outb( CShigh , Output ) ; 
  udelay( TWH ) ; // min=17us (t_wH) 
  outb( Clow , Output ) ;
  
  udelay( TSU ) ;  // (t_SU) 

  for( i = 0 ; i < 8 ; ++i ) { 

  if( inb( Input ) & BitW ) EV += BV ;
  BV /= 2 ;
  outb( CLKhigh , Output ) ;
  udelay( TTIO ) ;  // t_t(I/O)
  outb( Clow , Output ) ;
  udelay( TTIO ) ;
  
  }
  

  OutBufPtr = OutBuf ; // puts the pointer to the first character
  sprintf( OutBuf , "%d\n" , EV ) ; // creates the the string 
  

// copies the string to the user_space
  while( *OutBufPtr && length ) { 
  
    put_user( *(OutBufPtr++)  , buffer++ ) ; 
    bytes_read++ ;
    length-- ;
  
  }

  return bytes_read ; // returns the number of bytes read - otherwise nothing happens

}




// module declarations

static int Major;
struct file_operations Fops = {


  read: device_read, 
  open: device_open,
  release: device_release  
  

};


// initialize the module
int init_module()
{

  printk( "%s: Device driver for the universal A/D-converter by Frank Steiner.\n%s: Copyright by Mathias Paulsen (matpaulsen@gmx.de).\n" , DEVICE_NAME , DEVICE_NAME ) ; 
  printk( "%s: There is no warranty !!!\n" , DEVICE_NAME ) ;


  if( iottyS0 == 0 ) iottyS0 = 0x3f8 ; // set default io_address
  if( iottyS1 == 0 ) iottyS1 = 0x2f8 ;
  if( iottyS2 == 0 ) iottyS2 = 0x3e8 ;
  if( iottyS3 == 0 ) iottyS3 = 0x2e8 ;
  if( iolp0 == 0 ) iolp0 = 0x378 ;
  if( iolp1 == 0 ) iolp1 = 0x278 ;

  printk( "%s: using iottyS0=0x%x\n" , DEVICE_NAME , iottyS0 ) ;
  printk( "%s: using iottyS1=0x%x\n" , DEVICE_NAME , iottyS1 ) ;
  printk( "%s: using iottyS2=0x%x\n" , DEVICE_NAME , iottyS2 ) ;
  printk( "%s: using iottyS3=0x%x\n" , DEVICE_NAME , iottyS3 ) ;
  printk( "%s: using iolp0=0x%x\n"   , DEVICE_NAME , iolp0   ) ;
  printk( "%s: using iolp1=0x%x\n"   , DEVICE_NAME , iolp1   ) ;
  
 

  
  // Register the character device
   
  Major = register_chrdev(0,
			  DEVICE_NAME,
			  &Fops
			  );
  
  
  if (Major < 0) {
    printk ("%s: %s %s device failed with %d\n", DEVICE_NAME , 
            "Sorry, registering the", DEVICE_NAME ,
            Major);
    return Major;
  }

  
  printk("%s: Major is %d\n" ,DEVICE_NAME , Major ) ;


  return 0;
}

// cleanup module
void cleanup_module()
{
  

  // since unregister_chrdev returns void there is nothing else to do
 
  unregister_chrdev(Major, DEVICE_NAME);

 
}

MODULE_LICENSE("Copyright 2010 Mathias Paulsen (matpaulsen@gmx.de) GPL V3");
